
import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.io.*;

/**
 * This class pulls everything together and runs the game.  It is also
 * responsible for handling key presses and mouse clicks.
 *
 * @author Michael Yu
 */
public class LunarLanderGame {

    public static LunarLander lander;    // Lunar lander object
    public static Moon moon;             // The moon we're landing on!
    public static GameCanvas canvas;     // Game canvas
    public static double time;           // Current time
    public static int score;             // Current score
    public static DataRecorder recorder; // Data recorder for the lunar lander

    /**
     * The main method run the lunar lander game
     *
     * @param args is the command-line arguments
     */
    public static void main(String[] args) {
        loadSettingsFromFile("settings.txt");
        loadSettingsFromArgs(args);

        // Create event handler
        eventHandler = new EventHandler();

        // Create frame
        frame = new JFrame();
        frame.setTitle("Lunar Lander");

        // Create and add lander canvas
        canvas = new GameCanvas(trace);
        canvas.setPreferredSize(new Dimension(800, 600));
        canvas.setMode(GameCanvas.TITLE_SCREEN);
        frame.getContentPane().add(canvas, "Center");

        // Create and add btnStartOver button at the top
        JPanel buttons = createButtonPanel();
        frame.getContentPane().add(buttons, "South");

        // Setup frame
        frame.pack();
        frame.addKeyListener(eventHandler);
        frame.setFocusable(true);
        frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
        Dimension d = Toolkit.getDefaultToolkit().getScreenSize();
        frame.setLocation((d.width - frame.getWidth()) / 2,
                (d.height - frame.getHeight()) / 2);
        frame.setResizable(false);
        frame.setVisible(true);
        frame.requestFocusInWindow();

        // Create a redrawTimer and start it
        redrawTimer = new Timer((int) (DRAW_DT * 1000), eventHandler);
    }

    /**
     * Load the game settings from the given file
     *
     * @param filename is the file to load the settings from
     */
    private static void loadSettingsFromFile(String filename) {
        try {
            BufferedReader in = new BufferedReader(new FileReader(filename));
            String str = in.readLine();
            int numMessages = Integer.parseInt(str);
            gameOverMessages = new String[numMessages];

            for (int i = numMessages - 1; i >= 0; i--) {
                if ((str = in.readLine()) == null) {
                    throw new Exception("Invalid settings.txt file.");
                }

                gameOverMessages[i] = str;
            }

            in.close();

        }
        catch (IOException e) {
            System.err.println("Could not open settings.txt");
            System.exit(1);
        }
        catch (Exception e) {
            System.err.println("Invalid settings.txt");
            System.exit(1);
        }
    }

    /**
     * Process any command-line arguments
     *
     * @param args is the command-line arguments
     */
    private static void loadSettingsFromArgs(String[] args) {
        if (args.length == 1 &&
                (args[0].equals("-h") || args[0].equals("--help"))) {
            printUsage();
            System.exit(0);
        }

        for (int i = 0; i < args.length; i += 2) {
            if (args.length == i + 1) {
                printUsage();
                System.exit(1);
            }

            if (!setParameter(args[i], args[i + 1])) {
                printUsage();
                System.exit(1);
            }
        }
    }

    /**
     * Print out the available command-line arguments for the program
     */
    private static void printUsage() {
        System.out.println("LunarLanderGame 1.0, an enhanced clone of the classic Lunar " +
                "Lander arcade game.");
        System.out.println("");
        System.out.println("Use the arrow keys to rotate the lander and control engine " +
                "thrust.  The goal is to land gently on one of the landing " +
                "pads.  Your score depends on how gently you land, the size " +
                "of the landing pad, and the amount of fuel used.");
        System.out.println("");
        System.out.println("Usage: LunarLanderGame [OPTIONS] ...");
        System.out.println("");
        System.out.println("OPTIONS:");
        System.out.println(
                "  -auto  on/off        : turn on/off the autopilot (default off)");
        System.out.println(
                "  -turbo X             : set the simulation speed multiplier to X (default 1.0)");
        System.out.println(
                "  -trace on/off        : trace the flight path of the lander (default off)");
        System.out.println(
                "  -terrain random/flat/filename : generate a random, flat, or specified terrain (default random)");
        System.out.println(
                "  -vx X                : set the initial x-velocity to X (default 10.0)");
        System.out.println(
                "  -vy X                : set the initial y-velocity to X (default 0.0)");
        System.out.println(
                "  -angle X             : set the initial angle to X (default 0.0)");
        System.out.println(
                "  -gravity X           : set the acceleration from gravity to X (default 2.0)");
        // TODO: Don't hard-code default values
    }

    /**
     * Set the game parameters
     *
     * @param parameter is the parameter to set
     * @param value     is the value the parameter should be set to
     * @return true if a valid parameter and value were given, false otherwise
     */
    private static boolean setParameter(String parameter, String value) {
        boolean valid = true;

        if (parameter.equals("-auto")) {
            if (value.equals("on")) {
                auto = true;
            }
            else if (value.equals("off")) {
                auto = false;
            }
            else {
                valid = false;
            }
        }
        else if (parameter.equals("-turbo")) {
            try {
                turbo = Double.parseDouble(value);
            }
            catch (NumberFormatException nfx) {
                valid = false;
            }
        }
        else if (parameter.equals("-trace")) {
            if (value.equals("on")) {
                trace = true;
            }
            else if (value.equals("off")) {
                trace = false;
            }
            else {
                valid = false;
            }
        }
        else if (parameter.equals("-terrain")) {
            if (value.equals("random")) {
                flatTerrain = false;
            }
            else if (value.equals("flat")) {
                flatTerrain = true;
            }
            else {
                loadTerrain = true;
                valid = readTerrainFile(value);
            }
        }
        else if (parameter.equals("-vx")) {
            try {
                initVel =
                        new Vect2D(Double.parseDouble(value), initVel.getY());
            }
            catch (NumberFormatException nfx) {
                valid = false;
            }
        }
        else if (parameter.equals("-vy")) {
            try {
                initVel =
                        new Vect2D(initVel.getX(), Double.parseDouble(value));
            }
            catch (NumberFormatException nfx) {
                valid = false;
            }
        }
        else if (parameter.equals("-angle")) {
            try {
                initAngle = Double.parseDouble(value);
            }
            catch (NumberFormatException nfx) {
                valid = false;
            }
        }
        else if (parameter.equals("-gravity")) {
            try {
                gravity = Double.parseDouble(value);
            }
            catch (NumberFormatException nfx) {
                valid = false;
            }
        }
        else {
            valid = false;
        }

        return valid;
    }

    private static boolean readTerrainFile(String fileName) {
        
        BufferedReader reader = null;
        try {        
            reader = new BufferedReader(new FileReader(fileName));
            String line = readUsefulLine(reader);
            
            if (line == null) {
                return false;
            }
            
            String[] pos = line.split(",");
            if (pos.length == 1) {
                loadedStartX = Double.parseDouble(pos[0].trim());
            } else {
                loadedStartX = Double.parseDouble(pos[0].trim());
                loadedStartY = Double.parseDouble(pos[1].trim());
            }
            
            line = readUsefulLine(reader);
            if (line == null) {
                return false;
            }
            
            String[] heights = line.split(",");
            loadedHeightMap = new double[heights.length];
            for(int i=0; i<loadedHeightMap.length; i++) {
                loadedHeightMap[i] = Double.parseDouble(heights[i].trim());
            }
            
            line = readUsefulLine(reader);
            if (line == null) {
                return false;
            }
            
            String[] longPads = line.split(",");
            loadedLongPads = new int[longPads.length];
            for(int i=0; i<loadedLongPads.length; i++) {
                loadedLongPads[i] = Integer.parseInt(longPads[i].trim());
            }
            
            line = readUsefulLine(reader);
            if(line == null) {
                return false;
            }
            
            String[] shortPads = line.split(",");
            loadedShortPads = new int[shortPads.length];
            for(int i=0; i<loadedShortPads.length; i++) {
                loadedShortPads[i] = Integer.parseInt(shortPads[i].trim());
            }
            
            return loadedStartX != -1 && loadedHeightMap != null && 
            		loadedLongPads != null && loadedShortPads != null;
        } catch (Exception e) {
            System.err.println("Invalid terrain file " + fileName + "; " + e);
            return false;
        } finally {
            if (reader != null) {
                try {
                    reader.close();
                } catch (IOException e) {
                    System.err.println("Unable to close file: " + e);
                }
            }
        }
        
        
    }
    
    /**
     * Create the button panel
     */
    private static JPanel createButtonPanel() {
        JPanel buttons = new JPanel();

        btnStartOver = new JButton("Start Over");
        btnStartOver.addActionListener(eventHandler);
        btnStartOver.setEnabled(false);
        buttons.add(btnStartOver);

        btnGenerateRandomLandscape = new JButton("Generate Random Landscape");
        btnGenerateRandomLandscape.addActionListener(eventHandler);
        buttons.add(btnGenerateRandomLandscape);

        btnSaveFlightData = new JButton("Save Flight Data");
        btnSaveFlightData.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                JFileChooser fc = new JFileChooser((File) null);
                fc.showSaveDialog(frame);
                File file = fc.getSelectedFile();

                if (file != null) {
                    try {
                        recorder.dumpToFile(file.toString());
                    }
                    catch (IOException iox) {
                        JOptionPane.showMessageDialog(frame,
                                "Could not write file: " + iox.getMessage(),
                                "Error", JOptionPane.ERROR_MESSAGE);
                        return;
                    }
                    JOptionPane.showMessageDialog(frame, "Flight data successfully written to " +
                            file.toString() + ".", "File Written",
                            JOptionPane.INFORMATION_MESSAGE);
                }
            }
        });
        buttons.add(btnSaveFlightData);
        btnSaveFlightData.setEnabled(false);

        return buttons;
    }

    /**
     * Return the appropriate game over message for the current score
     *
     * @return the game over message based on the current score
     */
    public static String getGameOverMessage() {
        if (lander.getStatus() == LunarLander.STATUS_LANDED) {
            int index = (int) ((gameOverMessages.length - 1) *
                    Math.min(1.0, (double) score / 130000.0));
            return gameOverMessages[index];
        }
        else {
            return "CRASHED!";
        }
    }

    /**
     * Calculate the score if the lander lands safely with the given velocity
     * and fuel remaining.  This method assumes the lander has landed safely!
     *
     * @param velocity is the landing velocity
     * @param fuel     is the fuel remaining
     * @return the score
     */
    public static int calculateScore(Vect2D velocity, double fuel) {
        double vx = velocity.getX();
        double vy = velocity.getY();
        double fuelPercentage = fuel / LunarLander.MAX_FUEL;
        double vxPercentage = (LunarLander.SAFE_VELOCITY_X - Math.abs(vx)) /
                LunarLander.SAFE_VELOCITY_X;
        double vyPercentage = (LunarLander.SAFE_VELOCITY_Y - Math.abs(vy)) /
                LunarLander.SAFE_VELOCITY_Y;

        int score = (int) (40000 * (fuelPercentage - 0.5) +
                30000 * Math.pow(vxPercentage, 2) +
                30000 * Math.pow(vyPercentage, 2));
        score *= moon.getScoreMultiplier(lander.getPosition().getX());
        return score;
    }
    
    private static String readUsefulLine(BufferedReader reader) throws IOException {
        String line = reader.readLine();
        while(line != null && (line.trim().length() == 0 || line.charAt(0) == '#')) {
            line = reader.readLine();
        }
        return line;
    }

    /**
     * Reset the game
     */
    private static void reset() {
        verticalAuto = false;
        score = 0;
        time = 0;
        double initX = loadTerrain ? loadedStartX : Math.random() * WORLD_WIDTH;
        double initY = loadTerrain ? loadedStartY : 2.0/3.0 * WORLD_HEIGHT;
        lander = new LunarLander(new Vect2D(initX, initY), initVel, initAngle);
        recorder = new DataRecorder(lander);
        autopilot = new Autopilot(lander);
        verticalAutopilot = new VerticalAutopilot(lander);
        eventHandler.resetKeys();
        redrawTimer.restart();
        canvas.setMode(GameCanvas.IN_GAME);
        canvas.setDisplayMessage(null);
        canvas.zoomOut();
        canvas.repaint();
        frame.requestFocusInWindow();
        btnSaveFlightData.setEnabled(false);
        btnStartOver.setEnabled(true);
    }

    /**
     * Update the game score
     */
    private static void updateScore() {
        if (lander.landedSafely()) {
            score = calculateScore(lander.getVelocity(), lander.getFuel());
        }
        else {
            score = 0;
        }
    }

    private static JFrame frame;                        // Graphics window
    private static JButton btnStartOver;                // Start over button
    private static JButton btnGenerateRandomLandscape;  // Generate random landscape button
    private static JButton btnSaveFlightData;           // Save flight data button
    private static Timer redrawTimer;                   // Timer for screem redraws
    private static EventHandler eventHandler;           // Event handler
    private static double turbo = 1;                    // Turbo multiplier
    private static boolean auto = false;                // Full autopilot engaged?
    private static boolean verticalAuto = false;        // Vertical autopilot engaged?
    private static boolean trace = false;               // Trace the flight path?
    private static boolean flatTerrain = false;         // Use flat terrain instead of random?    
    private static String[] gameOverMessages;           // Gave over messages
    private static Vect2D initVel = new Vect2D(10, 0);  // meters/sec
    private static double initAngle = 0.0;              // rad
    private static double gravity = 2.0;                // meters/sec^2
    private static Autopilot autopilot;                 // Full autopilot
    private static VerticalAutopilot verticalAutopilot; // Vertical autopilot

    public static final double THROTTLE_SENSITIVITY = 2.5;
    public static final double DT = .01;                   // Simulation timestep
    public static final double DRAW_DT = .05;              // Draw timestep
    public static final double WORLD_WIDTH = 1000.0;       // meters
    public static final double WORLD_HEIGHT = 750.0;       // meters
    
    private static boolean loadTerrain = false;			// Use the supplied terrain
    private static double[] loadedHeightMap = null;		// height map for loaded terrain
    private static double loadedStartX = -1;			// starting x location for loaded terrain
    private static double loadedStartY = 2.0/3*WORLD_HEIGHT;	// starting y location for loaded terrain
    private static int[] loadedLongPads = null;
    private static int[] loadedShortPads = null;

    /**
     * EventHandler class handles key presses and button presses.  It also
     * runs the simulation.
     */
    private static class EventHandler implements ActionListener, KeyListener {

        /**
         * This method handles all action events (such as button presses and
         * redrawTimer alarms).
         *
         * @param e is the action event
         */
        public void actionPerformed(ActionEvent e) {
            redrawTimer.stop();

            // "Start over" button was pressed
            if (e.getSource() == btnStartOver) {
                reset();

            }
            // "Generate random landscape" button was pressed
            else if (e.getSource() == btnGenerateRandomLandscape) {
                Terrain terrain = 
                    loadTerrain ? TerrainGenerator.generateLoaded(WORLD_WIDTH, loadedHeightMap,
                            loadedLongPads, loadedShortPads) :
                    flatTerrain ? TerrainGenerator.generateFlat(
                        WORLD_WIDTH, WORLD_HEIGHT / 10.0) : 
                            	  TerrainGenerator.generateRandom(
                                WORLD_WIDTH, 10.0, 2 * WORLD_HEIGHT / 5, 6,
                                0.5);
                moon = new Moon(terrain, WORLD_WIDTH, WORLD_HEIGHT, gravity);
                reset();

            }
            // Redraw timer went off
            else if (e.getSource() == redrawTimer) { // Timer interrupt
                simulate();

                updateScore();
                canvas.repaint();

                // Restart the redraw timer
                if (lander.getStatus() == LunarLander.STATUS_FLYING) {
                    redrawTimer.restart();
                }
                else {
                    btnSaveFlightData.setEnabled(true);
                    canvas.setDisplayMessage(getGameOverMessage());
                    canvas.repaint();
                }
            }
        }

        /**
         * Run the simulation for one draw cycle
         */
        private void simulate() {
            // If autopilot is turned off, updated the throttle and angle of the
            // lunar lander based on which keys the user has pressed
            if (!auto) {
                // Check if UP is pressed
                if (keyUpPressed || keyUpTyped) {
                    keyUpTyped = false;
                    lander.setThrottle(lander.getThrottle() +
                            (THROTTLE_SENSITIVITY * DRAW_DT));
                }

                // Check if DOWN is pressed
                if (keyDownPressed || keyDownTyped) {
                    keyDownTyped = false;
                    lander.setThrottle(lander.getThrottle() -
                            (THROTTLE_SENSITIVITY * DRAW_DT));
                }

                // Check if LEFT is pressed
                if (keyLeftPressed || keyLeftTyped) {
                    keyLeftTyped = false;

                    double desiredAngle = lander.getAngle() +
                            LunarLander.TURN_RATE * DRAW_DT;

                    // This guarantees we can bring the lander back to the
                    // straight upright position
                    if (lander.getAngle() < 0 && desiredAngle > 0) {
                        lander.setDesiredAngle(0);
                    }
                    else {
                        lander.setDesiredAngle(desiredAngle);
                    }
                }

                // Check if RIGHT is pressed
                if (keyRightPressed || keyRightTyped) {
                    keyRightTyped = false;

                    double desiredAngle = lander.getAngle() -
                            LunarLander.TURN_RATE * DRAW_DT;

                    // This guarantees we can bring the lander back to the
                    // straight upright position
                    if (lander.getAngle() > 0 && desiredAngle < 0) {
                        lander.setDesiredAngle(0);
                    }
                    else {
                        lander.setDesiredAngle(desiredAngle);
                    }
                }
            }

            // Run the simulation
            for (int i = 0; i < turbo * DRAW_DT / DT; i++) {
                if (auto) {
                    autopilot.updateLanderInputs(time);
                }
                if (verticalAuto) {
                    verticalAutopilot.updateLanderInputs();
                }

                time += DT;
                lander.step(DT);
                recorder.record(DT);
            }
        }

        /**
         * This method is called when the user taps a key.  It updates the state
         * variables for the event handler
         *
         * @param e is the key event
         */
        public void keyTyped(KeyEvent e) {
            switch (e.getKeyCode()) {
                case KeyEvent.VK_UP:
                    keyUpPressed = true;
                    break;
                case KeyEvent.VK_DOWN:
                    keyDownPressed = true;
                    break;
                case KeyEvent.VK_LEFT:
                    keyLeftPressed = true;
                    break;
                case KeyEvent.VK_RIGHT:
                    keyRightPressed = true;
                    break;
                case KeyEvent.VK_SPACE:
                    verticalAutopilot.engageVerticalAutopilot();
                    verticalAuto = true;
                    break;
            }

            switch (e.getKeyChar()) {
                case 'w': // fall through
                case 'i':
                    keyUpPressed = keyUpTyped = true;
                    break;
                case 's': // fall through
                case 'k':
                    keyDownPressed = keyDownTyped = true;
                    break;
                case 'a': // fall through
                case 'j':
                    keyLeftPressed = keyLeftTyped = true;
                    break;
                case 'd': // fall through
                case 'l':
                    keyRightPressed = keyRightTyped = true;
                    break;
            }

        }

        /**
         * This method is called when the user holds down a key.  It updates the
         * state variables for the event handler
         *
         * @param e is the key event
         */
        public void keyPressed(KeyEvent e) {
            switch (e.getKeyCode()) {
                case KeyEvent.VK_UP:
                    keyUpPressed = keyUpTyped = true;
                    break;
                case KeyEvent.VK_DOWN:
                    keyDownPressed = keyDownTyped = true;
                    break;
                case KeyEvent.VK_LEFT:
                    keyLeftPressed = keyLeftTyped = true;
                    break;
                case KeyEvent.VK_RIGHT:
                    keyRightPressed = keyRightTyped = true;
                    break;
                case KeyEvent.VK_SPACE:
                    verticalAutopilot.engageVerticalAutopilot();
                    verticalAuto = true;
                    break;
            }

            switch (e.getKeyChar()) {
                case 'w': // fall through
                case 'i':
                    keyUpPressed = keyUpTyped = true;
                    break;
                case 's': // fall through
                case 'k':
                    keyDownPressed = keyDownTyped = true;
                    break;
                case 'a': // fall through
                case 'j':
                    keyLeftPressed = keyLeftTyped = true;
                    break;
                case 'd': // fall through
                case 'l':
                    keyRightPressed = keyRightTyped = true;
                    break;
            }
        }

        /**
         * This method is called when the user releases a key.  It updates the
         * state variables for the event handler
         *
         * @param e is the key event
         */
        public void keyReleased(KeyEvent e) {
            switch (e.getKeyCode()) {
                case KeyEvent.VK_UP:
                    keyUpPressed = false;
                    break;
                case KeyEvent.VK_DOWN:
                    keyDownPressed = false;
                    break;
                case KeyEvent.VK_LEFT:
                    keyLeftPressed = false;
                    break;
                case KeyEvent.VK_RIGHT:
                    keyRightPressed = false;
                    break;
            }

            switch (e.getKeyChar()) {
                case 'w': // fall through
                case 'i':
                    keyUpPressed = false;
                    break;
                case 's': // fall through
                case 'k':
                    keyDownPressed = false;
                    break;
                case 'a': // fall through
                case 'j':
                    keyLeftPressed = false;
                    break;
                case 'd': // fall through
                case 'l':
                    keyRightPressed = false;
                    break;
            }
        }

        /**
         * Reset the keys to "unpressed" state
         */
        public void resetKeys() {
            keyUpPressed = false;
            keyDownPressed = false;
            keyLeftPressed = false;
            keyRightPressed = false;
            keyUpTyped = false;
            keyDownTyped = false;
            keyLeftTyped = false;
            keyRightTyped = false;
        }

        boolean keyUpPressed;    // Is the user holding down the up arrow key?
        boolean keyDownPressed;  // Is the user holding down the down arrow key?
        boolean keyLeftPressed;  // Is the user holding down the left arrow key?
        boolean keyRightPressed; // Is the user holding down the right arrow key?
        boolean keyUpTyped;      // Did the user tap the up arrow key?
        boolean keyDownTyped;    // Did the user tap the down arrow key?
        boolean keyLeftTyped;    // Did the user tap the left arrow key?
        boolean keyRightTyped;   // Did the user tap the right arrow key?
    }
}