
import java.awt.Graphics;
import java.util.List;
import java.util.ArrayList;
import java.util.Iterator;

/**
 * A moon.  Each moon has a unique surface (terrain) for the lunar lander to
 * land on
 *
 * @author Michael Yu
 */
public class Moon {

    /**
     * Construct a celestial body with the given terrain
     *
     * @param terrain     is the surface shape of the moon
     * @param worldWidth  is the circumference of the moon
     * @param worldHeight is the max height still considered inside the moon's
     *                    realm
     * @param gravity     is the acceleration due to gravity on this moon
     */
    public Moon(Terrain terrain, double worldWidth, double worldHeight,
            double gravity) {
        this.terrain = terrain;
        this.worldWidth = worldWidth;
        this.worldHeight = worldHeight;
        this.gravity = gravity;
    }

    /**
     * Get the acceleration due to gravity on this moon
     *
     * @return the acceleration due to gravity on this moon
     */
    public double getGravity() {
        return gravity;
    }

    /**
     * Return the height of the terrain
     *
     * @param x is the x position
     * @return the height of the terrain at positions getX
     */
    public double getTerrainHeight(double x) {
        return terrain.getHeight(x);
    }

    /**
     * Is there a landing pad directly below?
     *
     * @param x is the x position
     * @return true if there is a landing pad at the given x position, false
     *         otherwise
     */
    public boolean isOverLandingPad(double x) {
        return terrain.isOverLandingPad(x);
    }

    /**
     * Get the score multiplier if the lunar lander landed on the surface of
     * this moon at position x
     *
     * @param x is the x position of the landing spot
     * @return the score multiplier for landing at position x
     */
    public int getScoreMultiplier(double x) {
        if (isOverLandingPad(x)) {
            return terrain.getEncompassingSegment(x).getMultiplier();
        }
        else {
            return 0;
        }
    }

    /**
     * Get the world height (maximum altitude that is considered within the
     * moon's region)
     *
     * @return the world height
     */
    public double getWorldHeight() {
        return worldHeight;
    }

    /**
     * Get the world width (the circumference of the moon)
     *
     * @return the world width
     */
    public double getWorldWidth() {
        return worldWidth;
    }

    /**
     * Get the terrain segment that contains the given x position
     *
     * @param x is the x position
     * @return the terrain segment that contains the given x position
     */
    public TerrainSegment getEncompassingTerrainSegment(double x) {
        return terrain.getEncompassingSegment(x);
    }

    /**
     * Get the closest landing pad (closest in terms of horizontal distance to
     * the left of the x-coordinate).
     *
     * @param x is the x-coordinate
     * @return the left-closest landing pad
     */
    public TerrainSegment getLeftClosestLandingPad(double x) {
        List landingPads = getLandingPads();
        TerrainSegment closestSegmentSoFar = null;
        double closestDistanceSoFar = Double.POSITIVE_INFINITY;

        for (Iterator it = landingPads.iterator(); it.hasNext();) {
            TerrainSegment s = (TerrainSegment) it.next();
            double landingPadX = (s.getRightEndPoint().getX() +
                    s.getLeftEndPoint().getX()) / 2.0;
            double distance = (landingPadX > x) ?
                    worldWidth - (landingPadX - x) : x - landingPadX;
            if (closestDistanceSoFar > distance) {
                closestDistanceSoFar = distance;
                closestSegmentSoFar = s;
            }
        }

        return closestSegmentSoFar;
    }

    /**
     * Get the closest landing pad (closest in terms of horizontal distance to
     * the right of the x-coordinate).
     *
     * @param x is the x-coordinate
     * @return the right-closest landing pad
     */
    public TerrainSegment getRightClosestLandingPad(double x) {
        List landingPads = getLandingPads();
        TerrainSegment closestSegmentSoFar = null;
        double closestDistanceSoFar = Double.POSITIVE_INFINITY;

        for (Iterator it = landingPads.iterator(); it.hasNext();) {
            TerrainSegment s = (TerrainSegment) it.next();
            double landingPadX = (s.getRightEndPoint().getX() +
                    s.getLeftEndPoint().getX()) / 2.0;
            double distance = (landingPadX < x) ?
                    worldWidth - (x - landingPadX) : landingPadX - x;
            if (closestDistanceSoFar > distance) {
                closestDistanceSoFar = distance;
                closestSegmentSoFar = s;
            }
        }

        return closestSegmentSoFar;
    }

    /**
     * @return a list of all the landing pads on the moon.  The landing pads are
     *         returned in order from left to right
     */
    public List getLandingPads() {
        ArrayList landingPads = new ArrayList();
        for (Iterator it = terrain.getSegments().iterator(); it.hasNext();) {
            TerrainSegment s = (TerrainSegment) it.next();
            if (s.isLandingPad()) {
                landingPads.add(s);
            }
        }
        return landingPads;
    }

    /**
     * Get the terrain
     *
     * @return the terrain
     */
    public Terrain getTerrain() {
        return terrain;
    }

    /**
     * Draw the moon's terrain
     *
     * @param g is the graphics context
     */
    public void draw(Graphics g) {
        terrain.draw(g);
    }

    private Terrain terrain;
    private double worldWidth;
    private double worldHeight;
    private double gravity;
}
