import java.awt.*;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.util.ArrayList;
import java.util.Iterator;
import javax.swing.Timer;

/**
 * A wrapping terrain / landscape.
 *
 * @author Michael Yu
 */
public class Terrain {

    /**
     * Constructs a new Terrain object, initializing the segment heights to the
     * heights given in the height map.
     *
     * @param heightMap  is an array of heights specifying the heights of the
     *                   terrain segments
     * @param worldWidth is the width of the entire terrain
     */
    Terrain(double[] heightMap, double worldWidth) {
        this.worldWidth = worldWidth;
        this.segments = new ArrayList(heightMap.length);
        double segmentWidth = worldWidth / heightMap.length;

        // Create the terrain segments
        for (int i = 0; i < heightMap.length; i++) {
            int iNext = (i + 1) % heightMap.length;
            Vect2D left = new Vect2D(i * segmentWidth, heightMap[i]);
            Vect2D right = new Vect2D((i + 1) * segmentWidth, heightMap[iNext]);
            segments.add(i, new TerrainSegment(left, right));
        }

        // Setup redrawTimer to blink the score multipliers
        timer = new Timer(900, new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                timer.stop();
                TerrainSegment.setMultiplierDisplayed(
                        !TerrainSegment.isMultiplierDisplayed());
                timer.start();
            }
        });
        timer.start();
    }

    /**
     * Find the index to the segment that covers the given x position
     *
     * @param x is the x position
     * @return the index to the segment that encompasses positionX
     */
    public int getEncompassingSegmentIndex(double x) {
        while (x < 0) {
            x += worldWidth;
        }

        while (x >= worldWidth) {
            x -= worldWidth;
        }

        for (int i = 0; i < segments.size(); i++) {
            TerrainSegment segment = (TerrainSegment) segments.get(i);
            if (segment.getLeftEndPoint().getX() <= x &&
                    x < segment.getRightEndPoint().getX()) {
                return i;
            }
        }

        // We should never get here.  This should really throw an exception,
        // but the students haven't learned exceptions yet.
        System.err.println("Terrain.getEncompassingSegmentIndex: ERROR: location " + x + " is not found on the Moon!");
        System.exit(1);
        return -1;
    }

    /**
     * Find the segment that covers the given x position
     *
     * @param x is the x position
     * @return the terrain segment that encompasses positions getX
     */
    public TerrainSegment getEncompassingSegment(double x) {
        return (TerrainSegment) segments.get(getEncompassingSegmentIndex(x));
    }

    /**
     * Return the height of the terrain
     *
     * @param x is the x position
     * @return the height of the terrain at positions getX
     */
    public double getHeight(double x) {
        while (x < 0) {
            x += worldWidth;
        }

        while (x >= worldWidth) {
            x -= worldWidth;
        }
        
        TerrainSegment segment = getEncompassingSegment(x);
        Vect2D left = segment.getLeftEndPoint();
        Vect2D right = segment.getRightEndPoint();
        double rise = right.getY() - left.getY();
        double run = right.getX() - left.getX();
        return left.getY() + (x - left.getX()) * rise / run;
    }

    /**
     * Is there a landing pad directly below?
     *
     * @param x is the x position
     * @return true if there is a landing pad at the given x position, false
     *         otherwise
     */
    public boolean isOverLandingPad(double x) {
        return getEncompassingSegment(x).isLandingPad();
    }

    /**
     * Draw the terrain
     *
     * @param g is the graphics context
     */
    public void draw(Graphics g) {
        for (Iterator it = segments.iterator(); it.hasNext();) {
            TerrainSegment segment = (TerrainSegment) it.next();
            segment.draw(g);
        }
    }

    /**
     * Get all the terrain segments
     *
     * @return a list of all the terrain segments
     */
    public ArrayList getSegments() {
        return segments;
    }

    private double worldWidth;    // the width of the world
    private ArrayList segments;   // all the terrain segments
    private Timer timer;          // redrawTimer used to blink the score multipliers
}
